#   Copyright  2020 Atos Spain SA. All rights reserved.
 
#   This file is part of EASIER AI.
 
#   EASIER AI is free software: you can redistribute it and/or modify it under the terms of Apache License, either version 2 of the License, or
#   (at your option) any later version.
 
#   THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT ANY WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
#   BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT,
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#   WHETHER IN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#   See  LICENSE file for full license information  in the project root.
from tensorflow.keras import Sequential, optimizers
from tensorflow.keras.layers import Dense, Dropout

import tensorflow.keras.models

from phased_lstm_keras.PhasedLSTM import PhasedLSTM


class Predictor:
    learning_rate = 0.001

    decay_lr = 0

    loss_fn = 'mean_squared_error'

    def __init__(self, output_activation, ft_range,  lr=0.001):
        self.model = None
        self.output_activation = output_activation
        self.ft_range = ft_range
        self.learning_rate = lr

    def predict(self, input):
        return self.model.predict(input)

    def classify(self, input):
        return self.model.predict_proba(input), self.model.predict_classes(input)

    def load_model(self, path):
        self.model = keras.models.load_model(path, custom_objects={"PhasedLSTM": PhasedLSTM})
        self.model._make_predict_function()

    def get_model(self, shape, num_forecasts, inner_activation='tanh'):
        model = Sequential()
        # Shape = (Samples, Timesteps, Features)
        model.add(PhasedLSTM(units=128, input_shape=shape,
                             return_sequences=False, activation=inner_activation))
        model.add(Dropout(0.2))

        model.add(Dense(units=num_forecasts, activation=self.output_activation))

        opt = optimizers.Adagrad(lr=self.learning_rate, decay=self.decay_lr)
        # opt = optimizers.rmsprop(lr=0.01)
        model.compile(optimizer=opt, loss=self.loss_fn, metrics=['mae'])
        model.summary()
        return model



#   Copyright  2020 Atos Spain SA. All rights reserved.
 
#   This file is part of EASIER AI.
 
#   EASIER AI is free software: you can redistribute it and/or modify it under the terms of Apache License, either version 2 of the License, or
#   (at your option) any later version.
 
#   THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT ANY WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
#   BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT,
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#   WHETHER IN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#   See  LICENSE file for full license information  in the project root.
import logging
from logging.handlers import TimedRotatingFileHandler
import datetime as dt
import time
import inspect
import os
import sys
from copy import copy

import logstash

import json

import common_functions.constants as constants

class ConsoleFormatter(logging.Formatter):
    
    def __init__(self):
        fmt = '%(asctime)s [%(levelname)-7s - %(filename)s] %(message)s %(add_info)s'
        datefmt = '%Y-%m-%d %H:%M:%S'
        logging.Formatter.__init__(self, fmt, datefmt)

    def format(self, record):
        record_copy = copy(record)
        record_copy.msg = record.msg
        record_copy.filename = record.source_file
        if hasattr(record, 'additional_info'):
            record_copy.add_info = '-> '+ str(record.additional_info)
        else:
            record_copy.add_info = ''
        return super().format(record_copy)

class Logger:

    def __init__(self, service, source_file):
        self.service = service
        self.source_file = source_file
        logstash_host = os.getenv('LOGSTASH_HOST')
        logstash_port = os.getenv('LOGSTASH_PORT')
        self.init_log_file(source_file, logstash_host, logstash_port)

    def lineno(self):
        f = inspect.currentframe()
        filename = os.path.normcase(f.f_code.co_filename)
        # Traverse the stack until the function has been called by a different file, then get that line number
        while filename == __file__:
            f = f.f_back
            filename = os.path.normcase(f.f_code.co_filename)

        return f.f_lineno

    def init_log_file(self, source_file, logstash_host=None, logstash_port=None):
        streamHandler = logging.StreamHandler(sys.stdout)
        streamHandler.setFormatter(ConsoleFormatter())
        self.logger = logging.getLogger(source_file)
        if logstash_host and logstash_port:
            self.logger.addHandler(logstash.TCPLogstashHandler(logstash_host, int(logstash_port), version=1))
        self.logger.addHandler(streamHandler)
        # By default, level is ERROR. If not recognized, level is also ERROR
        logging_level = os.getenv('LOGGING_LEVEL', 'ERROR').upper()
        if logging_level == 'WARN' or logging_level == 'WARNING':
            self.logger.setLevel(logging.WARNING)
        elif logging_level == 'DEBUG':
            self.logger.setLevel(logging.DEBUG)
        elif logging_level == 'INFO':
            self.logger.setLevel(logging.INFO)
        else:
            self.logger.setLevel(logging.ERROR)
        
    def generate_extra_doc(self, level, additional_info=None):
        doc = {
            'service': self.service,
            'source_file': self.source_file + ':' + str(self.lineno()),
            '@timestamp': dt.datetime.now().isoformat(),
            'debug_level': level
        }
        if additional_info:
            doc['additional_info'] = additional_info
        return doc

    def debug(self, message, additional_info=None):
        self.logger.debug(message, extra=self.generate_extra_doc(constants.DEBUG, additional_info))

    def info(self, message, additional_info=None):
        self.logger.info(message, extra=self.generate_extra_doc(constants.INFO, additional_info))

    def warning(self, message, additional_info=None):
        self.logger.warning(message, extra=self.generate_extra_doc(constants.WARNING, additional_info))

    def error(self, message, additional_info=None):
        self.logger.error(message, extra=self.generate_extra_doc(constants.ERROR, additional_info))

    def perf(self, message, additional_info=None):
        self.logger.info(message, extra=self.generate_extra_doc(constants.PERFORMANCE, additional_info))


#   Copyright  2020 Atos Spain SA. All rights reserved.
 
#   This file is part of EASIER AI.
 
#   EASIER AI is free software: you can redistribute it and/or modify it under the terms of Apache License, either version 2 of the License, or
#   (at your option) any later version.
 
#   THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT ANY WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING 
#   BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT,
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
#   WHETHER IN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#   See  LICENSE file for full license information  in the project root.
from tensorflow.keras import Sequential, optimizers
from tensorflow.keras.layers import Dense, Flatten
import tensorflow.keras.models


class Predictor:
    learning_rate = 0.01

    decay_lr = 0

    loss_fn = 'categorical_crossentropy'

    def __init__(self, output_activation, ft_range, lr=0.001, loss = 'categorical_crossentropy'):
        self.model = None
        self.output_activation = output_activation
        self.ft_range = ft_range
        self.learning_rate = lr
        self.loss_fn = loss

    def predict(self, input):
        return self.model.predict(input)

    def classify(self, input):
        return self.model.predict_proba(input), self.model.predict_classes(input)

    def load_model(self, path):
        self.model = tensorflow.keras.models.load_model(path)
        self.model._make_predict_function()

    def get_model(self, shape, num_classes, inner_activation='relu'):
        model = Sequential()
        # Input arrays of shape (*, layers[1])
        # Output = arrays of shape (*, layers[1] * 16)
        model.add(Dense(units=int(64), input_shape=shape, activation=inner_activation))
        model.add(Dense(units=int(64), activation=inner_activation))
        # model.add(Dropout(0.2))
        model.add(Flatten())
        model.add(Dense(units=num_classes, activation=self.output_activation))

        # opt = optimizers.Adagrad(lr=self.learning_rate, epsilon=None, decay=self.decay_lr)
        # opt = optimizers.RMSprop(lr=0.001)
        model.compile(optimizer="adam", loss=self.loss_fn, metrics=['acc'])
        model.summary()
        self.model = model
        return model

    def get_model_configurable(self, shape, num_classes, inner_activation='relu', num_dense_layers=1, num_units_per_layer=[64], loss='categorical_crossentropy', metrics=['acc','categorical_accuracy']):
        model = Sequential()
        # Input arrays of shape (*, layers[1])
        # Output = arrays of shape (*, layers[1] * 16)
        model.add(Dense(units=int(num_units_per_layer[0]), input_shape=shape, activation=inner_activation))
        for i in range(1,num_dense_layers):
            model.add(Dense(units=int(num_units_per_layer[i]), activation=inner_activation))
        
        model.add(Flatten())
        model.add(Dense(units=num_classes, activation=self.output_activation))

        # opt = optimizers.Adagrad(lr=self.learning_rate, epsilon=None, decay=self.decay_lr)
        # opt = optimizers.RMSprop(lr=self.learning_rate)
        model.compile(loss=loss, metrics=metrics)
        model.summary()
        self.model = model
        return model

